/*
 * data.h - Typed data containers
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#ifndef DATA_H
#define DATA_H


#include <stdint.h>
#include <stdio.h>


/* including tree.h would cause circular dependency via param.h:PARAM */
struct _qdisc;
struct _class;
struct _filter;
struct _police;
/* ditto for op.h */
struct _op;
/* ditto for field.h */
struct _field;
struct _field_root;


typedef enum {
    dt_none,	/* no data */
    dt_unum,	/* unsigned integer (don't trust double to preserve a u32) */
    dt_fnum,	/* floating-point number */
    dt_string,	/* string (malloc'ed) */
    dt_rate,	/* rate (bits per second) */
    dt_size,	/* size (Bytes) */
    dt_prate,	/* rate (packets per second) */
    dt_psize,	/* size (packets) */
    dt_time,	/* time (seconds) */
    dt_device,	/* device (internal use only) */
    dt_qdisc,	/* queuing discipline */
    dt_class,	/* class */
    dt_filter,	/* filter */
    dt_tunnel,	/* tunnel (internal use only) */
    dt_police,	/* policer */
    dt_list,	/* list of data elements */
    dt_field,	/* field */
    dt_field_root,
		/* field root */
    dt_decision,/* generalized policing decision */
} DATA_TYPE;

#define DT_NON_NUMERIC	dt_qdisc /* limit for numeric types */


struct _data;


typedef enum {
    dr_continue,	/* continue/unspec */
    dr_class,		/* pass/ok */
    dr_drop,		/* drop/shot */
    dr_reclassify,	/* reclassify */
} DECISION_RESULT;

typedef struct _data_list {
    struct _data *ref;		/* pointer to actual data */
    struct _data_list *next;	/* next list element */
} DATA_LIST;

typedef struct _data {
    DATA_TYPE type;
    struct _op *op;
	/* if non-NULL, u is invalid and the expression is variable */
    union {
	uint32_t unum;		/* dt_unum */
	double fnum;		/* dt_fnum, dt_p?rate, dt_p?size, dt_time */
	char *string;		/* dt_string */
	struct _device *device;	/* dt_device */
	struct _qdisc *qdisc;	/* dt_qdisc */
	struct _class *class;	/* dt_class */
	struct _filter *filter;	/* dt_filter */
	struct _tunnel *tunnel;	/* dt_tunnel */
	struct _police *police;	/* dt_police */
	DATA_LIST *list;	/* dt_list */
	struct _field *field;	/* dt_field */
	struct _field_root *field_root;
				/* dt_field_root */
	struct {
	    DECISION_RESULT result;
	    const struct _class *class;
	} decision;		/* dt_decision */
    } u;
} DATA;


DATA data_none(void);
DATA data_unum(uint32_t in);
DATA data_fnum(double in);
DATA data_string(const char *in);
DATA data_device(struct _device *in);
DATA data_qdisc(struct _qdisc *in);
DATA data_class(struct _class *in);
DATA data_decision(DECISION_RESULT result,struct _class *class);
DATA data_filter(struct _filter *in);
DATA data_tunnel(struct _tunnel *in);
DATA data_police(struct _police*in);
DATA data_list(DATA_LIST *list);
DATA_LIST *data_list_element(DATA *ref);
DATA data_field(struct _field *field);
DATA data_field_root(struct _field_root *field_root);

int all_unum(DATA a,DATA b);

const char *type_name(DATA_TYPE type);

DATA data_convert(DATA in,DATA_TYPE type);

/*
 * unit.type must be dt_none, dt_p?rate, dt_p?size, or dt_time
 */

DATA data_add_unit(DATA in,DATA unit);

DATA data_clone(DATA d);
int data_equal(DATA a,DATA b);

//#define DONT_DESTROY_DATA
#ifdef DONT_DESTROY_DATA
#define data_destroy_1(x) ((void) &x)
#define data_destroy(x) ((void) &x)
#else
void data_destroy_1(DATA d);
void data_destroy(DATA d);
#endif

void print_data(FILE *file,DATA d);

#endif /* DATA_H */
